import pytest
import sqlalchemy as sa

from validatable.inference import prepare_column_name


@pytest.mark.parametrize(
    ("column", "column_name"),
    [
        (sa.Column("samename", sa.String), "samename"),
        (sa.Column(sa.String), "None"),
    ],
    ids=["same name", "None"],
)
def test_correct_column_name(column: sa.Column, column_name: str):

    result = prepare_column_name(column, column_name)
    assert result.name == column_name


@pytest.mark.parametrize(
    ("column", "column_name"),
    [
        (sa.Column("different", sa.String), "name"),
    ],
    ids=["different name"],
)
def test_incorrect_column_name(column: sa.Column, column_name: str):

    with pytest.raises(ValueError):
        prepare_column_name(column, column_name)



import datetime as dt
import enum
import os
import pathlib
import random
import uuid
from decimal import Decimal
from ipaddress import (
    IPv4Address,
    IPv4Interface,
    IPv4Network,
    IPv6Address,
    IPv6Interface,
    IPv6Network,
    ip_address,
    ip_interface,
    ip_network,
)

from faker import Faker
from pydantic import (
    UUID1,
    UUID3,
    UUID4,
    UUID5,
    AnyHttpUrl,
    AnyUrl,
    EmailStr,
    Field,
    HttpUrl,
    IPvAnyAddress,
    IPvAnyInterface,
    IPvAnyNetwork,
    NameEmail,
    NegativeFloat,
    NegativeInt,
    PositiveFloat,
    PositiveInt,
    StrictBool,
    conbytes,
    condecimal,
    confloat,
    conint,
    constr,
    stricturl,
)

from validatable.main import BaseTable, MetaData

Faker.seed(0)
faker = Faker()


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ UUID


class ModelUUID1(BaseTable, metadata=MetaData()):
    test: UUID1 = Field(default_factory=uuid.uuid1)


class ModelUUID3(BaseTable, metadata=MetaData()):
    test: UUID3 = Field(
        default_factory=lambda: uuid.uuid3(uuid.uuid4(), "name")
    )


class ModelUUID4(BaseTable, metadata=MetaData()):
    test: UUID4 = Field(default_factory=uuid.uuid4)


class ModelUUID5(BaseTable, metadata=MetaData()):
    test: UUID5 = Field(
        default_factory=lambda: uuid.uuid5(uuid.uuid4(), "name")
    )


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Number


class ModelInt(BaseTable, metadata=MetaData()):

    test: int = Field(
        default_factory=lambda: random.randint(-2147483648, 2147483647)
    )


class ModelConInt(BaseTable, metadata=MetaData()):
    test: conint() = Field(  # type: ignore[valid-type]
        default_factory=lambda: random.randint(-2147483648, 2147483647)
    )


class ModelFloat(BaseTable, metadata=MetaData()):
    test: float = Field(
        default_factory=lambda: random.uniform(
            -3.402823466e38, 1.175494351e-38
        )
    )


class ModelConFloat(BaseTable, metadata=MetaData()):
    test: confloat() = Field(  # type: ignore[valid-type]
        default_factory=lambda: random.uniform(
            -3.402823466e38, 1.175494351e-38
        )
    )


class ModelDecimal(BaseTable, metadata=MetaData()):
    test: Decimal = Field(
        default_factory=lambda: Decimal(
            f"{random.uniform(-1000000000, 1000000000):.10f}"
        )
    )


class ModelConDecimal(BaseTable, metadata=MetaData()):
    test: condecimal() = Field(  # type: ignore[valid-type]
        default_factory=lambda: Decimal(
            f"{random.uniform(-1000000000, 1000000000):.10f}"
        )  # noqa E501
    )


class ModelPositiveInt(BaseTable, metadata=MetaData()):
    test: PositiveInt = Field(
        default_factory=lambda: random.randint(0, 2147483647)
    )


class ModelNegativeInt(BaseTable, metadata=MetaData()):
    test: NegativeInt = Field(
        default_factory=lambda: random.randint(-2147483648, 0)
    )


class ModelPositiveFloat(BaseTable, metadata=MetaData()):
    test: PositiveFloat = Field(
        default_factory=lambda: random.uniform(0, 1.175494351e-38)
    )


class ModelNegativeFloat(BaseTable, metadata=MetaData()):
    test: NegativeFloat = Field(
        default_factory=lambda: random.uniform(-3.402823466e38, 0)
    )


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ String Bytes


class ModelStr(BaseTable, metadata=MetaData()):
    test: str = Field(default_factory=faker.sentence, sa_primary_key=True)


class ModelConStr(BaseTable, metadata=MetaData()):
    test: constr(max_length=10) = Field(  # type: ignore[valid-type]
        default_factory=lambda: f"{faker.sentence()}"[:10]
    )


class ModelEmailStr(BaseTable, metadata=MetaData()):
    test: EmailStr = Field(default_factory=faker.email)


class ModelNameEmail(BaseTable, metadata=MetaData()):
    # NameEmail do not allow points "." in validation. Example: 'Mr.', 'Jr.'
    # https://github.com/samuelcolvin/pydantic/issues/2955
    test: NameEmail = Field(
        default_factory=lambda: NameEmail(
            name=faker.name().replace(".", ""), email=faker.email()
        )
    )


class ModelPath(BaseTable, metadata=MetaData()):
    test: pathlib.Path = Field(
        default_factory=lambda: pathlib.PosixPath(
            faker.file_path(depth=random.randint(1, 5))
        )
    )


class ModelBytes(BaseTable, metadata=MetaData()):
    test: bytes = Field(default_factory=lambda: os.urandom(10))


class ModelConBytes(BaseTable, metadata=MetaData()):
    test: conbytes(max_length=10) = Field(  # type: ignore[valid-type]
        default_factory=lambda: os.urandom(10)
    )


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Network


class ModelIPv4Address(BaseTable, metadata=MetaData()):
    test: IPv4Address = Field(default_factory=lambda: ip_address(faker.ipv4()))


class ModelIPv4Interface(BaseTable, metadata=MetaData()):
    test: IPv4Interface = Field(
        default_factory=lambda: ip_interface(faker.ipv4(network=True))
    )


class ModelIPv4Network(BaseTable, metadata=MetaData()):
    test: IPv4Network = Field(
        default_factory=lambda: ip_network(faker.ipv4(network=True))
    )


class ModelIPv6Address(BaseTable, metadata=MetaData()):
    test: IPv6Address = Field(default_factory=lambda: ip_address(faker.ipv6()))


class ModelIPv6Interface(BaseTable, metadata=MetaData()):
    test: IPv6Interface = Field(
        default_factory=lambda: ip_interface(faker.ipv6(network=True))
    )


class ModelIPv6Network(BaseTable, metadata=MetaData()):
    test: IPv6Network = Field(
        default_factory=lambda: ip_network(faker.ipv6(network=True))
    )


class ModelIPvAnyAddress(BaseTable, metadata=MetaData()):
    test: IPvAnyAddress = Field(
        default_factory=lambda: ip_address(
            faker.ipv6() if random.choice([True, False]) else faker.ipv4()
        )
    )


class ModelIPvAnyInterface(BaseTable, metadata=MetaData()):
    test: IPvAnyInterface = Field(
        default_factory=lambda: ip_interface(
            faker.ipv6(network=True)
            if random.choice([True, False])
            else faker.ipv4(network=True)
        )
    )


class ModelIPvAnyNetwork(BaseTable, metadata=MetaData()):
    test: IPvAnyNetwork = Field(
        default_factory=lambda: ip_network(
            faker.ipv6(network=True)
            if random.choice([True, False])
            else faker.ipv4(network=True)
        )
    )


class ModelHttpUrl(BaseTable, metadata=MetaData()):
    test: HttpUrl = Field(default_factory=faker.url)


class ModelAnyUrl(BaseTable, metadata=MetaData()):
    test: AnyUrl = Field(default_factory=faker.url)


class ModelAnyHttpUrl(BaseTable, metadata=MetaData()):
    test: AnyHttpUrl = Field(default_factory=faker.url)


class ModelStrictUrl(BaseTable, metadata=MetaData()):
    test: stricturl() = Field(  # type: ignore[valid-type]
        default_factory=faker.url
    )


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ datetime


class ModelDatetime(BaseTable, metadata=MetaData()):
    test: dt.datetime = Field(default_factory=faker.date_time_between)


class ModelDate(BaseTable, metadata=MetaData()):
    test: dt.date = Field(default_factory=faker.date_between)


class ModelTime(BaseTable, metadata=MetaData()):
    test: dt.time = Field(default_factory=faker.time_object)


class ModelTimedelta(BaseTable, metadata=MetaData()):
    test: dt.timedelta = Field(
        default_factory=lambda: dt.timedelta(
            seconds=random.randint(0, 999999999)
        )
    )


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Enum


class CaseEnum(enum.Enum):
    a: str = "a1"
    b: str = "b2"
    c: str = "c3"
    d: str = "d4"
    e: str = "e5"
    f: str = "f6"
    g: str = "g7"
    h: str = "h8"
    i: str = "i9"
    j: str = "j0"


class CaseIntEnum(enum.IntEnum):
    a: int = enum.auto()
    b: int = enum.auto()
    c: int = enum.auto()
    d: int = enum.auto()
    e: int = enum.auto()
    f: int = enum.auto()
    g: int = enum.auto()
    h: int = enum.auto()
    i: int = enum.auto()
    j: int = enum.auto()


class ModelEnum(BaseTable, metadata=MetaData()):
    test: CaseEnum = Field(
        default_factory=lambda: getattr(
            CaseEnum, random.choice(list(CaseEnum.__members__.keys()))
        )
    )


class ModelIntEnum(BaseTable, metadata=MetaData()):
    test: CaseIntEnum = Field(
        default_factory=lambda: getattr(
            CaseIntEnum, random.choice(list(CaseIntEnum.__members__.keys()))
        )
    )


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Boolean


class ModelBool(BaseTable, metadata=MetaData()):
    test: bool = Field(default_factory=lambda: random.choice((True, False)))


class ModelStrictBool(BaseTable, metadata=MetaData()):
    test: StrictBool = Field(
        default_factory=lambda: random.choice((True, False))
    )



import os

import pytest

from validatable import MetaData, create_engine

NUM_REPETITIONS = int(os.getenv("N") or 1)
DATABASE = os.getenv("DB") or "sqlite"
metadata = MetaData()


@pytest.fixture(scope="session")
def engine():

    if DATABASE == "sqlite":
        yield create_engine("sqlite:///:memory:")
    if DATABASE == "postgresql":
        yield create_engine(
            "postgresql://validatable:password@localhost:5432/db"
        )
    if DATABASE == "mariadb":
        yield create_engine(
            "mariadb+pymysql://validatable:password@localhost:3306/db"
        )


@pytest.fixture(scope="function")
def conn(engine):
    metadata.create_all(engine)
    with engine.connect() as conn:
        yield conn
    metadata.drop_all(engine)


class MakeConnection(object):
    def __init__(self, engine):
        self.engine = engine

    def __call__(self, model):
        self.metadata = model.metadata
        self.metadata.create_all(self.engine)
        self.conn = self.engine.connect()
        return self.conn

    def close(self):
        self.conn.close()
        self.metadata.drop_all(self.engine)


@pytest.fixture()
def make_conn(engine):
    make_conn = MakeConnection(engine)
    yield make_conn
    make_conn.close()



import pytest
import sqlalchemy as sa
from sqlalchemy import ForeignKey

from validatable.fields import Field
from validatable.main import BaseTable


@pytest.mark.parametrize(
    "kwarg",
    [
        {"sa_primary_key": True},
        {"sa_nullable": False},
        {"sa_index": True},
        {"sa_unique": True},
        {"sa_args": []},
    ],
)
def test_sa_arguments(kwarg):
    field = Field(**kwarg)
    key = tuple(kwarg.keys())[0]
    value = tuple(kwarg.values())[0]
    assert field.extra[key] == value


def test_model_primary_key():

    metadata = sa.MetaData()

    class Item(BaseTable, metadata=metadata):
        id: int = Field(sa_primary_key=True)
        name: str = Field("")

    assert Item.c.id.primary_key


def test_model_nullable():

    metadata = sa.MetaData()

    class Item(BaseTable, metadata=metadata):
        id: int = Field(sa_primary_key=True)
        name: str = Field("", sa_nullable=False)

    assert not Item.c.name.nullable


def test_model_index():

    metadata = sa.MetaData()

    class Item(BaseTable, metadata=metadata):
        id: int = Field(sa_primary_key=True)
        name: str = Field("", sa_index=True)

    assert Item.c.name.index


def test_model_unique():

    metadata = sa.MetaData()

    class Item(BaseTable, metadata=metadata):
        id: int = Field(sa_primary_key=True)
        name: str = Field("", sa_unique=True)

    assert Item.c.name.unique


def test_model_foreign_key():

    metadata = sa.MetaData()

    fk = ForeignKey("test.id")

    class Item(BaseTable, metadata=metadata):
        id: int = Field(sa_primary_key=True)
        name: str = Field("", sa_foreign_key=fk)

    assert len(Item.c.name.foreign_keys) == 1
    assert fk in Item.c.name.foreign_keys


def test_model_foreign_key_with_fk():

    metadata = sa.MetaData()

    class Item(BaseTable, metadata=metadata):
        id: int = Field(sa_primary_key=True)
        name: str = Field(
            "", sa_fk=sa.ForeignKey("test.id", ondelete="CASCADE")
        )

    fk = tuple(Item.c.name.foreign_keys)[0]

    assert len(Item.c.name.foreign_keys) == 1
    assert fk.target_fullname == "test.id"



from typing import Any, List, Set, Tuple
from uuid import UUID

import pytest
from sqlalchemy.dialects import mysql, postgresql, sqlite
from sqlalchemy.schema import Column, CreateColumn

from validatable.generic_types import GUID, AutoJson, AutoString, SLBigInteger


def test_guid_type_DDL_postgresql():
    c = Column("c", GUID)
    t_ddl = CreateColumn(c)

    postgresql_ddl = str(t_ddl.compile(dialect=postgresql.dialect()))
    _, postgresql_type = postgresql_ddl.split()

    assert "UUID" == postgresql_type


def test_guid_type_DDL_sqlite():
    c = Column("c", GUID)
    t_ddl = CreateColumn(c)

    sqlite_ddl = str(t_ddl.compile(dialect=sqlite.dialect()))
    _, sqlite_type = sqlite_ddl.split()

    assert "BINARY(16)" == sqlite_type


def test_guid_type_DDL_mysql():
    c = Column("c", GUID)
    t_ddl = CreateColumn(c)

    mysql_ddl = str(t_ddl.compile(dialect=mysql.dialect()))
    _, mysql_type = mysql_ddl.split()

    assert "BINARY(16)" == mysql_type


def test_guid_python_type():
    assert GUID().python_type == UUID


@pytest.mark.parametrize(
    ["value", "dialect", "expected"],
    (
        (
            [
                UUID("08e6ff8e253545d584558a50a57667c5"),
                "postgresql",
                UUID("08e6ff8e253545d584558a50a57667c5"),
            ],
            [
                None,
                "postgresql",
                None,
            ],
            [
                UUID("08e6ff8e253545d584558a50a57667c5"),
                "Any other",
                UUID("08e6ff8e253545d584558a50a57667c5").bytes,
            ],
            [
                None,
                "Any other",
                None,
            ],
        )
    ),
)
def test_guid_process_bind_param(value, dialect, expected):
    DialectMock = type("DialectMock", (object,), {"name": dialect})
    guid = GUID()
    result = guid.process_bind_param(value, DialectMock)
    assert result == expected


@pytest.mark.parametrize(
    ["value", "dialect", "expected"],
    (
        (
            [
                UUID("08e6ff8e253545d584558a50a57667c5"),
                "postgresql",
                UUID("08e6ff8e253545d584558a50a57667c5"),
            ],
            [
                None,
                "postgresql",
                None,
            ],
            [
                UUID("08e6ff8e253545d584558a50a57667c5").bytes,
                "Any other",
                UUID("08e6ff8e253545d584558a50a57667c5"),
            ],
            [
                None,
                "Any other",
                None,
            ],
        )
    ),
)
def test_guid_process_result_value(value, dialect, expected):
    DialectMock = type("DialectMock", (object,), {"name": dialect})
    guid = GUID()
    result = guid.process_result_value(value, DialectMock)
    assert result == expected


def test_autostring_python_type():
    assert AutoString().python_type == str


def test_autostring_type_DDL_postgresql():
    c = Column("c", AutoString)
    t_ddl = CreateColumn(c)

    postgresql_ddl = str(t_ddl.compile(dialect=postgresql.dialect()))
    _, postgresql_type = postgresql_ddl.split()

    assert "VARCHAR" == postgresql_type


def test_autostring_type_DDL_sqlite():
    c = Column("c", AutoString)
    t_ddl = CreateColumn(c)

    sqlite_ddl = str(t_ddl.compile(dialect=sqlite.dialect()))
    _, sqlite_type = sqlite_ddl.split()

    assert "VARCHAR" == sqlite_type


def test_autostring_type_DDL_mysql():
    c = Column("c", AutoString)
    t_ddl = CreateColumn(c)

    mysql_ddl = str(t_ddl.compile(dialect=mysql.dialect()))
    _, mysql_type = mysql_ddl.split()

    assert "VARCHAR(512)" == mysql_type


@pytest.mark.parametrize(
    ["value", "dialect", "expected"],
    (
        (
            [
                "SomeString",
                "Any",
                "SomeString",
            ],
            [
                None,
                "Any",
                None,
            ],
            [
                1,
                "Any",
                "1",
            ],
        )
    ),
)
def test_autostring_process_bind_param(value, dialect, expected):
    DialectMock = type("DialectMock", (object,), {"name": dialect})
    auto_string = AutoString()
    result = auto_string.process_bind_param(value, DialectMock)
    assert result == expected


@pytest.mark.parametrize(
    ["value", "dialect", "expected"],
    (
        (
            [
                "SomeString",
                "Any",
                "SomeString",
            ],
            [
                None,
                "Any",
                None,
            ],
        )
    ),
)
def test_autostring_process_result_value(value, dialect, expected):
    DialectMock = type("DialectMock", (object,), {"name": dialect})
    auto_string = AutoString()
    result = auto_string.process_result_value(value, DialectMock)
    assert result == expected


def test_slbiginteger_type_DDL_sqlite():
    c = Column("c", SLBigInteger)
    t_ddl = CreateColumn(c)

    sqlite_ddl = str(t_ddl.compile(dialect=sqlite.dialect()))
    _, sqlite_type = sqlite_ddl.split()

    assert "INTEGER" == sqlite_type


@pytest.mark.parametrize(
    "dialect, expected_type",
    [(sqlite, "INTEGER"), (mysql, "BIGINT"), (postgresql, "BIGINT")],
    ids=["sqlite", "mysql", "postgresql"],
)
def test_slbiginteger_type_DDL_not_sqlite(dialect, expected_type):
    c = Column("c", SLBigInteger)
    t_ddl = CreateColumn(c)

    dialect_ddl = str(t_ddl.compile(dialect=dialect.dialect()))
    _, dialect_type = dialect_ddl.split()

    assert dialect_type == expected_type


@pytest.mark.parametrize(
    "type_",
    [List, Any, Set, Tuple, list, set, tuple],
)
def test_autojson_python_type(type_):
    assert AutoJson(python_type=type_).python_type == type_



from collections import deque
from typing import Deque, List, Set, Tuple

from validatable import BaseTable, MetaData


def test_list_type():
    class ListCase(BaseTable, metadata=MetaData()):
        name: list = [1, 2, 3]


def test_typing_list_type():
    class ListCase(BaseTable, metadata=MetaData()):
        name: List


def test_typing_list_type_with_subscription():
    class ListCase(BaseTable, metadata=MetaData()):
        name: List[int]


def test_set_type():
    class SetCase(BaseTable, metadata=MetaData()):
        name: set = [1, 2, 3]


def test_typing_set_type():
    class SetCase(BaseTable, metadata=MetaData()):
        name: Set


def test_typing_set_type_with_subscription():
    class SetCase(BaseTable, metadata=MetaData()):
        name: Set[int]


def test_tuple_type():
    class TupleCase(BaseTable, metadata=MetaData()):
        name: tuple = [1, 2, 3]


def test_typing_tuple_type():
    class TupleCase(BaseTable, metadata=MetaData()):
        name: Tuple


def test_typing_tuple_type_with_subscription():
    class TupleCase(BaseTable, metadata=MetaData()):
        name: Tuple[int]


def test_deque_type():
    class DequeCase(BaseTable, metadata=MetaData()):
        name: deque


def test_typing_deque_type():
    class DequeCase(BaseTable, metadata=MetaData()):
        name: Deque


def test_typing_deque_with_subscription():
    class DequeCase(BaseTable, metadata=MetaData()):
        name: Deque[int]


def test_type_tuple(make_conn):
    class Model(BaseTable, metadata=MetaData()):
        ft: Tuple = ("a", "b", 1, 1.2)
        f: tuple = ("a", "b", 1, 1.2)
        f_sub: Tuple[int, ...] = (1, 2, 3, 4)

    instance = Model()
    create = Model.t.insert().values(instance.dict())
    read = Model.t.select()
    conn = make_conn(Model)
    conn.execute(create)
    result = conn.execute(read)
    data = result.fetchone()
    m = Model.construct(data)
    assert m == instance


def test_type_set(make_conn):
    class Model(BaseTable, metadata=MetaData()):
        ft: Set = {"a", "b", 1, 1.2}
        f: set = {"a", "b", 1, 1.2}
        f_sub: Set[int] = {1, 2, 3, 4}

    instance = Model()
    create = Model.t.insert().values(instance.dict())
    read = Model.t.select()
    conn = make_conn(Model)
    conn.execute(create)
    result = conn.execute(read)
    data = result.fetchone()
    m = Model.construct(data)
    assert m == instance


def test_type_list(make_conn):
    class Model(BaseTable, metadata=MetaData()):
        ft: List = ["a", "b", 1, 1.2]
        f: list = ["a", "b", 1, 1.2]
        f_sub: List[int] = [1, 2, 3, 4]

    instance = Model()
    create = Model.t.insert().values(instance.dict())
    read = Model.t.select()
    conn = make_conn(Model)
    conn.execute(create)
    result = conn.execute(read)
    data = result.fetchone()
    m = Model.construct(data)
    assert m == instance



from typing import Any, List, Optional

from pydantic.fields import FieldInfo, Undefined
from pydantic.typing import NoArgAnyCallable
from sqlalchemy import ForeignKey


def Field(
    default: Any = Undefined,
    *,
    default_factory: Optional[NoArgAnyCallable] = None,
    alias: str = None,
    title: str = None,
    description: str = None,
    const: bool = None,
    gt: float = None,
    ge: float = None,
    lt: float = None,
    le: float = None,
    multiple_of: float = None,
    min_items: int = None,
    max_items: int = None,
    min_length: int = None,
    max_length: int = None,
    allow_mutation: bool = True,
    regex: str = None,
    sa_primary_key: Optional[bool] = False,
    sa_pk: Optional[bool] = False,
    sa_nullable: Optional[bool] = None,
    sa_index: Optional[bool] = None,
    sa_unique: Optional[bool] = None,
    sa_args: List[Any] = None,
    sa_foreign_key: Optional[ForeignKey] = None,
    sa_fk: Optional[ForeignKey] = None,
    **extra: Any,
) -> Any:
    extra["sa_primary_key"] = sa_primary_key
    extra["sa_pk"] = sa_pk
    extra["sa_nullable"] = sa_nullable
    extra["sa_index"] = sa_index
    extra["sa_unique"] = sa_unique
    extra["sa_args"] = sa_args or []
    extra["sa_foreign_key"] = sa_foreign_key
    extra["sa_fk"] = sa_fk

    field_info = FieldInfo(
        default,
        default_factory=default_factory,
        alias=alias,
        title=title,
        description=description,
        const=const,
        gt=gt,
        ge=ge,
        lt=lt,
        le=le,
        multiple_of=multiple_of,
        min_items=min_items,
        max_items=max_items,
        min_length=min_length,
        max_length=max_length,
        allow_mutation=allow_mutation,
        regex=regex,
        **extra,
    )
    field_info._validate()
    return field_info



from typing import Type

import pytest
from conftest import NUM_REPETITIONS
from factories import (
    ModelAnyHttpUrl,
    ModelAnyUrl,
    ModelBool,
    ModelBytes,
    ModelConBytes,
    ModelConDecimal,
    ModelConFloat,
    ModelConInt,
    ModelConStr,
    ModelDate,
    ModelDatetime,
    ModelDecimal,
    ModelEmailStr,
    ModelEnum,
    ModelFloat,
    ModelHttpUrl,
    ModelInt,
    ModelIntEnum,
    ModelIPv4Address,
    ModelIPv4Interface,
    ModelIPv4Network,
    ModelIPv6Address,
    ModelIPv6Interface,
    ModelIPv6Network,
    ModelIPvAnyAddress,
    ModelIPvAnyInterface,
    ModelIPvAnyNetwork,
    ModelNameEmail,
    ModelNegativeFloat,
    ModelNegativeInt,
    ModelPath,
    ModelPositiveFloat,
    ModelPositiveInt,
    ModelStr,
    ModelStrictBool,
    ModelStrictUrl,
    ModelTime,
    ModelTimedelta,
    ModelUUID1,
    ModelUUID3,
    ModelUUID4,
    ModelUUID5,
)

from validatable import BaseTable


def make_params(*Models, repeat: int = NUM_REPETITIONS):
    param = []
    ids = []
    for Model in Models:
        param.extend([(Model, Model()) for n in range(repeat)])
        ids.extend([Model.__name__.replace("Model", "")] * repeat)
    return ids, param


ids, params = make_params(
    ModelInt,
    ModelUUID1,
    ModelUUID3,
    ModelUUID4,
    ModelUUID5,
    ModelConInt,
    ModelFloat,
    ModelConFloat,
    ModelDecimal,
    ModelConDecimal,
    ModelPositiveInt,
    ModelPositiveFloat,
    ModelNegativeFloat,
    ModelNegativeInt,
    ModelStr,
    ModelConStr,
    ModelBytes,
    ModelConBytes,
    ModelEmailStr,
    ModelNameEmail,
    ModelPath,
    ModelIPv4Address,
    ModelIPv4Interface,
    ModelIPv4Network,
    ModelIPv6Address,
    ModelIPv6Interface,
    ModelIPv6Network,
    ModelIPvAnyAddress,
    ModelIPvAnyInterface,
    ModelIPvAnyNetwork,
    ModelDate,
    ModelDatetime,
    ModelTime,
    ModelTimedelta,
    ModelEnum,
    ModelIntEnum,
    ModelHttpUrl,
    ModelAnyUrl,
    ModelAnyHttpUrl,
    ModelStrictUrl,
    ModelBool,
    ModelStrictBool,
)


@pytest.mark.parametrize(
    "Model, instance",
    params[::NUM_REPETITIONS],
    ids=[i.strip() for i in ids[::NUM_REPETITIONS]],
)
def test_factories(Model, instance):
    assert Model(**instance.dict()) == instance


@pytest.mark.filterwarnings(
    r"ignore:Dialect sqlite\+pysqlite does \*not\* "
    r"support Decimal objects natively"
)
@pytest.mark.parametrize("Model, instance", params, ids=ids)
def test_type_(Model: Type[BaseTable], instance: BaseTable, make_conn):
    create = Model.t.insert().values(instance.dict())  # type: ignore
    read = Model.t.select().where(  # type: ignore[union-attr]
        Model.c.test == instance.test  # type: ignore[attr-defined]
    )
    conn = make_conn(Model)
    conn.execute(create)
    result = conn.execute(read)
    data = result.fetchone()

    m = Model.construct(**data)
    assert m == instance



"""
The main module provides the definition of Validatable BaseTable.

It defines ValidatableMetaclass, which is a customization of BaseTable
creation that includes the instantiation of a SQLAlchemy Table object.

BaseTable class extends Pydantic BaseModel to include SQLAlchemy Table
instance methods in the class interface.

"""
from typing import Any, Dict, List, Optional, Set

from pydantic import BaseModel
from pydantic.main import ModelMetaclass
from sqlalchemy import Table
from sqlalchemy.sql.base import ImmutableColumnCollection
from sqlalchemy.sql.schema import MetaData

from .inference import get_table


class ValidatableMetaclass(ModelMetaclass):
    """Extends ModelMetaclass to include SQLAlchemy Table logic."""

    @classmethod
    def __prepare__(mcls, name, bases, *, metadata=None, **kwargs):
        """Set metadata before the evaluation of the class body."""
        if isinstance(metadata, MetaData):
            return {"__sa_metadata__": metadata, "__create_table__": True}

        elif metadata is None:

            if bases and hasattr(bases[0], "__sa_metadata__"):
                return {
                    "__sa_metadata__": bases[0].__sa_metadata__,
                    "__create_table__": bases[0].__create_table__,
                }
            else:
                return {}

        else:
            raise TypeError("metadata must be a sqlalquemy metadata")

    def __new__(mcls, name, bases, namespace, metadata=None, **kwargs):
        """Control the BaseTable definition."""
        table = namespace.get("__sa_table__")
        if isinstance(table, Table):
            return super().__new__(mcls, name, bases, namespace, **kwargs)

        metadata = namespace.get("__sa_metadata__", None)
        if metadata is None:
            namespace["__sa_metadata__"] = namespace.pop("metadata", None)
            namespace["__create_table__"] = namespace.get(
                "__create_table__", True
            )
            return super().__new__(mcls, name, bases, namespace, **kwargs)

        tablename = namespace.get("__sa_tablename__", name.lower())
        namespace["__sa_tablename__"] = tablename

        table_args = namespace.get("__sa_table_args__", [])
        namespace["__sa_table_args__"] = table_args

        table_kwargs = namespace.get("__sa_table_kwargs__", {})
        namespace["__sa_table_kwargs__"] = table_kwargs

        exclude = namespace.get("__sa_exclude__")

        cls = super().__new__(mcls, name, bases, namespace, **kwargs)

        if cls.__create_table__:
            cls.__create_table__ = False
            cls.__sa_table__ = get_table(
                tablename,
                metadata,
                cls.__fields__,
                table_args,
                table_kwargs,
                exclude,
            )
        else:
            cls.__sa_table__ = None
            cls.__sa_metadata__ = None
            cls.__sa_table_args__ = []
            cls.__sa_table_kwargs__ = {}
            cls.__sa_exclude__ = None
        return cls

    @property
    def c(cls) -> ImmutableColumnCollection:
        """Return the collection of columns."""
        return cls.__sa_table__.c  # type: ignore[attr-defined]

    @property
    def t(cls) -> Optional[Table]:
        """Return the table."""
        return cls.__sa_table__  # type: ignore[attr-defined]

    @property
    def metadata(cls) -> Optional[MetaData]:
        """Return the metadata instance."""
        return cls.__sa_metadata__  # type: ignore[attr-defined]


class BaseTable(BaseModel, metaclass=ValidatableMetaclass):
    """Extends BaseModel to include SQLAlchemy Table construction."""

    __sa_table__: Optional[Table]
    __sa_metadata__: Optional[MetaData]
    __sa_table_args__: List[Any]
    __sa_table_kwargs__: Dict[str, Any]
    __sa_exclude__: Optional[Set[str]] = None


class Validatable(BaseModel, metaclass=ValidatableMetaclass):
    """Extends BaseModel to include SQLAlchemy Table construction."""

    __sa_table__: Optional[Table]
    __sa_metadata__: Optional[MetaData]
    __sa_table_args__: List[Any]
    __sa_table_kwargs__: Dict[str, Any]
    __sa_exclude__: Optional[Set[str]] = None



import enum
import uuid
from decimal import Decimal
from typing import Dict, Optional

import pytest
from pydantic import (
    UUID1,
    UUID3,
    UUID4,
    UUID5,
    BaseModel,
    ConstrainedDecimal,
    ConstrainedFloat,
    ConstrainedInt,
    NegativeFloat,
    NegativeInt,
    NonNegativeFloat,
    NonNegativeInt,
    NonPositiveFloat,
    NonPositiveInt,
    PositiveFloat,
    PositiveInt,
    StrictFloat,
    StrictInt,
    condecimal,
    conint,
)
from pydantic.fields import ModelField
from pydantic.types import ConstrainedNumberMeta, JsonMeta

from validatable.type_dispatch import Dispatch, get_sql_type


class CaseEnum(enum.Enum):
    a: int = 1
    b: int = 2
    c: str = "c"


MAX_LENGTH: int = 10


class ModelCaseInt(BaseModel):

    python_int: int
    con_int: conint()  # type: ignore
    ccon_int: ConstrainedInt
    p_int: PositiveInt
    n_int: NegativeInt
    np_int: NonPositiveInt
    nn_int: NonNegativeInt
    s_int: StrictInt
    opython_int: Optional[int]
    ocon_int: Optional[conint()]  # type: ignore
    occon_int: Optional[ConstrainedInt]
    op_int: Optional[PositiveInt]
    on_int: Optional[NegativeInt]
    onp_int: Optional[NonPositiveInt]
    onn_int: Optional[NonNegativeInt]
    os_int: Optional[StrictInt]


class ModelCaseFloat(BaseModel):

    python_float: ConstrainedFloat
    p_float: PositiveFloat
    n_float: NegativeFloat
    np_float: NonPositiveFloat
    nn_float: NonNegativeFloat
    s_float: StrictFloat
    opython_float: Optional[ConstrainedFloat]
    op_float: Optional[PositiveFloat]
    on_float: Optional[NegativeFloat]
    onp_float: Optional[NonPositiveFloat]
    onn_float: Optional[NonNegativeFloat]
    os_float: Optional[StrictFloat]


class ModelCaseUUID(BaseModel):

    python_uuid: uuid.UUID
    uuid1: UUID1
    uuid3: UUID3
    uuid4: UUID4
    uuid5: UUID5
    opython_uuid: Optional[uuid.UUID]
    ouuid1: Optional[UUID1]
    ouuid3: Optional[UUID3]
    ouuid4: Optional[UUID4]
    ouuid5: Optional[UUID5]


class ModelCaseDecimal(BaseModel):
    python_decimal: Decimal
    con_decimal: condecimal(max_digits=10)  # type: ignore
    c_decimal: ConstrainedDecimal
    opython_decimal: Optional[Decimal]
    ocon_decimal: Optional[condecimal(max_digits=10)]  # type: ignore
    oc_decimal: Optional[ConstrainedDecimal]


class ModelCaseTypes(
    ModelCaseUUID, ModelCaseFloat, ModelCaseInt, ModelCaseDecimal
):
    ...


@pytest.fixture(scope="function")
def dispatch():
    @Dispatch
    def dispatch(m, *args, **kwargs):
        raise TypeError(f"Oh, No! {m.type_}, {m.outer_type_}")

    @dispatch.dispatcher
    def _(m: ModelField, *args, dispatch: Dispatch = None, **kwargs):

        if m.outer_type_ in dispatch._funcs:
            func = dispatch._funcs.get(m.outer_type_)
            return func(m, *args, **kwargs)

        meta_func = dispatch._funcs.get(m.outer_type_.__class__)

        if meta_func is None:
            return dispatch._base(m)
        else:
            func = meta_func(m, dispatch=dispatch)
            return func(m, *args, **kwargs)

    @dispatch.register(ConstrainedNumberMeta, JsonMeta, type)
    def _(m: ModelField, *args, dispatch: Dispatch = None, **kwargs):
        type_ = m.outer_type_
        if type_.__mro__[1] in dispatch._funcs:
            return dispatch._funcs[type_.__mro__[1]]

    yield dispatch


@pytest.mark.parametrize(
    "field",
    ModelCaseUUID.__fields__.values(),
    ids=ModelCaseUUID.__fields__.keys(),
)
def test_dispatch_uuid(field, dispatch):
    m = field

    @dispatch.register(uuid.UUID, UUID1, UUID3, UUID4, UUID5)
    def _(m):
        return "uuid"

    assert dispatch(m) == "uuid"


@pytest.mark.parametrize(
    "field",
    ModelCaseInt.__fields__.values(),
    ids=ModelCaseInt.__fields__.keys(),
)
def test_dispatch_int(field, dispatch):
    m = field

    @dispatch.register(
        int,
        PositiveInt,
        NegativeInt,
        NonPositiveInt,
        NonNegativeInt,
        StrictInt,
        ConstrainedInt,
    )
    def _(m):
        return "int"

    assert dispatch(m) == "int"


@pytest.mark.parametrize(
    "field",
    ModelCaseFloat.__fields__.values(),
    ids=ModelCaseFloat.__fields__.keys(),
)
def test_dispatch_float(field, dispatch):
    m = field

    @dispatch.register(
        float,
        ConstrainedFloat,
        PositiveFloat,
        NegativeFloat,
        NonPositiveFloat,
        NonNegativeFloat,
        StrictFloat,
    )
    def _(m):
        return "float"

    assert dispatch(m) == "float"


@pytest.mark.parametrize(
    "field",
    ModelCaseDecimal.__fields__.values(),
    ids=ModelCaseDecimal.__fields__.keys(),
)
def test_dispatch_decimal(field, dispatch):
    m = field

    @dispatch.register(
        Decimal,
        ConstrainedDecimal,
    )
    def _(m):
        return "decimal"

    assert dispatch(m) == "decimal"


@pytest.mark.parametrize(
    ("key", "field"),
    ModelCaseTypes.__fields__.items(),
    ids=ModelCaseTypes.__fields__.keys(),
)
def test_dispatch_types(field, key, dispatch):
    m = field

    @dispatch.register(uuid.UUID, UUID1, UUID3, UUID4, UUID5)
    def _(m):
        return "uuid"

    @dispatch.register(
        float,
        ConstrainedFloat,
        PositiveFloat,
        NegativeFloat,
        NonPositiveFloat,
        NonNegativeFloat,
        StrictFloat,
    )
    def _(m):
        return "float"

    @dispatch.register(
        int,
        PositiveInt,
        NegativeInt,
        NonPositiveInt,
        NonNegativeInt,
        StrictInt,
        ConstrainedInt,
    )
    def _(m):
        return "int"

    @dispatch.register(
        Decimal,
        ConstrainedDecimal,
    )
    def _(m):
        return "decimal"

    assert dispatch(m) in key


def test_wrong_type():
    class BaseWrong(BaseModel):
        test: Dict

    with pytest.raises(TypeError):
        get_sql_type(BaseWrong.__fields__["test"])



from functools import partial

from sqlalchemy.engine import create_engine as sa_create_engine

create_engine = partial(
    sa_create_engine,
    json_deserializer=lambda x: x,
    json_serializer=lambda x: x,
)

# import json
# from pydantic.json import pydantic_encoder
# create_engine = partial(
#     sa_create_engine,
#     json_deserializer=lambda x: x,
#     json_serializer=lambda obj: json.dumps(obj, default=pydantic_encoder),
# )



import sys

if sys.version_info < (3, 7):
    from typing import GenericMeta, TupleMeta

    typing_meta = (GenericMeta, TupleMeta)

    def get_type(t):
        return t._gorg.mro()[1]


elif sys.version_info < (3, 9):
    from typing import _GenericAlias  # type: ignore[attr-defined]
    from typing import _VariadicGenericAlias  # type: ignore[attr-defined]

    typing_meta = (_GenericAlias, _VariadicGenericAlias)

    def get_type(t):
        return t.__origin__


else:
    from typing import _GenericAlias  # type: ignore[attr-defined]
    from typing import _SpecialGenericAlias  # type: ignore[attr-defined]
    from typing import _TupleType  # type: ignore[attr-defined]

    typing_meta = (_GenericAlias, _SpecialGenericAlias, _TupleType)

    def get_type(t):
        return t.__origin__



from typing import List

from pydantic import parse_obj_as

from validatable import BaseTable, Field, MetaData


def test_nullable_column():
    class NotNull(BaseTable, metadata=MetaData()):
        not_null: str

    table = NotNull.__sa_table__
    m = NotNull.__fields__["not_null"]

    assert m.required
    assert not table.c.not_null.nullable


def test_nullable_column_elipsis():
    class NotNull(BaseTable, metadata=MetaData()):
        id: int = Field(sa_pk=True)
        not_null: str = Field(...)

    table = NotNull.__sa_table__
    m = NotNull.__fields__["not_null"]

    assert m.required
    assert not table.c.not_null.nullable


def test_nullable_column_only_sql():
    class NotNull(BaseTable, metadata=MetaData()):
        not_null: str = Field(None, sa_nullable=False)

    table = NotNull.__sa_table__
    m = NotNull.__fields__["not_null"]

    assert not table.c.not_null.nullable
    assert not m.required


def test_nullable_column_with_pk():
    class NotNull(BaseTable, metadata=MetaData()):
        not_null: int = Field(None, sa_pk=True)

    table = NotNull.__sa_table__
    m = NotNull.__fields__["not_null"]

    assert table.c.not_null.nullable is None
    assert not m.required


def test_nullable_column_with_pk_integration(make_conn):
    class NotNull(BaseTable, metadata=MetaData()):
        id: int = Field(None, sa_pk=True)
        test: int = Field(..., sa_unique=True)

    data = [NotNull(test=n).dict() for n in (1, 2, 3)]
    conn = make_conn(NotNull)
    conn.execute(NotNull.t.insert().values(data))
    result = conn.execute(NotNull.t.select())
    db_data = result.fetchall()

    not_nulls = parse_obj_as(List[NotNull], db_data)

    assert len(db_data) == 3
    assert len(not_nulls) == 3
    assert all(n.id is not None for n in not_nulls)


def test_nullable_column_with_pk_integration_sa_nullable_true(make_conn):
    class NotNull(BaseTable, metadata=MetaData()):
        id: int = Field(..., sa_pk=True, sa_nullable=True)
        test: int = Field(..., sa_unique=True)

    data = [NotNull(test=n, id=n).dict() for n in (1, 2, 3)]
    conn = make_conn(NotNull)
    conn.execute(NotNull.t.insert().values(data))
    result = conn.execute(NotNull.t.select())
    db_data = result.fetchall()

    not_nulls = parse_obj_as(List[NotNull], db_data)

    assert len(db_data) == 3
    assert len(not_nulls) == 3



from typing import Any, Dict, List, Optional, Set, Tuple, Union

import sqlalchemy as sa
from pydantic.fields import ModelField, UndefinedType

from .type_dispatch import get_sql_type


def prepare_column_name(column: sa.Column, column_name: str) -> sa.Column:
    if column.name == column_name:
        return column

    if column.name is None:
        column.name = column_name
        return column

    raise ValueError(
        "Column name must be equal to field name, or field alias, or None"
    )


def primary_key_kwargs(col_kwargs: Dict[str, Any]):
    pk = col_kwargs.pop("pk", False)
    return col_kwargs.get("primary_key") or pk


def nullable_kwargs(
    col_kwargs: Dict[str, Any], required: Union[bool, UndefinedType], pk: bool
):
    nullable = col_kwargs.get("nullable")
    if pk:
        return nullable
    return nullable if nullable is not None else not required


def get_sa_args_kwargs(m: ModelField) -> Tuple[Any, Dict[str, Any]]:
    keys = tuple(m.field_info.extra.keys())
    col_kwargs = {
        k[3:]: m.field_info.extra.pop(k) for k in keys if k.startswith("sa_")
    }
    col_kwargs["primary_key"] = primary_key_kwargs(col_kwargs)
    col_kwargs["nullable"] = nullable_kwargs(
        col_kwargs, m.required, col_kwargs["primary_key"]
    )

    args = col_kwargs.pop("args", [])
    fk = col_kwargs.pop("fk", None)
    fk = col_kwargs.pop("foreign_key", None) or fk
    args.append(fk)
    return args, col_kwargs


def get_column(m: ModelField) -> sa.Column:
    args, col_kwargs = get_sa_args_kwargs(m)
    column = col_kwargs.pop("column", None)

    if isinstance(column, sa.Column):
        return prepare_column_name(column, m.alias)

    column_type = col_kwargs.pop("type", None) or col_kwargs.pop("type_", None)

    if column_type:
        return sa.Column(m.alias, column_type, *args, **col_kwargs)

    sa_type = get_sql_type(m)
    return sa.Column(m.alias, sa_type, *args, **col_kwargs)


def is_model_field(v: Any) -> bool:
    return hasattr(v, "__class__") and isinstance(v, ModelField)


def get_table(
    name: str,
    metadata: sa.MetaData,
    fields: Dict[str, Any],
    table_args: List[str],
    table_kwargs: Dict[str, Any],
    exclude: Optional[Set[str]] = None,
) -> sa.Table:

    exclude = exclude or set()
    columns = [
        get_column(v)
        for k, v in fields.items()
        if k not in exclude and is_model_field(v)
    ]
    return sa.Table(name, metadata, *columns, *table_args, **table_kwargs)



import uuid
from random import randint

import pytest
import sqlalchemy as sa
from conftest import metadata
from faker import Faker

from validatable import UUID4, BaseTable, Field, MetaData, Validatable

Faker.seed(0)
faker = Faker()


@pytest.fixture(scope="function")
def conn(engine):
    metadata.create_all(engine)
    with engine.connect() as conn:
        yield conn
    metadata.drop_all(engine)


class Base(BaseTable):
    id: UUID4 = Field(default_factory=uuid.uuid4, sa_primary_key=True)
    metadata = metadata


class ModelUpdateDelete(Base):
    num: int = 0


def test_metadata_property():

    meta = sa.MetaData()

    class ModelCase(BaseTable, metadata=meta):
        id: int = Field(sa_primary_key=True)

    class Base(BaseTable):
        metadata = meta
        id: int = Field(sa_primary_key=True)

    class TableFromBase(Base):
        ...

    class Table(BaseTable):
        __sa_metadata__ = meta
        id: int = Field(sa_primary_key=True)

    assert TableFromBase.metadata == meta
    assert ModelCase.metadata == meta
    assert Table.metadata == meta


def test_table_declaration():

    table = sa.Table(
        "modelcase",
        metadata,
        sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
    )

    class ModelCase(BaseTable):

        __sa_table__ = table
        id: int = None

    assert table == ModelCase.__sa_table__


def test_raise_for_wrong_metadata():

    with pytest.raises(TypeError):

        class ModelCase(BaseTable, metadata="NotMetaData"):
            id: int = None


def test_database_update(conn):
    model = ModelUpdateDelete()
    insert = ModelUpdateDelete.t.insert().values(model.dict())

    update = (
        ModelUpdateDelete.t.update()
        .where(ModelUpdateDelete.c.id == model.id)
        .values(num=5)
    )
    one = ModelUpdateDelete.t.select().where(
        ModelUpdateDelete.c.id == model.id
    )

    conn.execute(insert)

    conn.execute(update)
    result_updated = conn.execute(one)
    data_updated = result_updated.fetchone()
    model_updated = ModelUpdateDelete.parse_obj(data_updated)

    assert model_updated.id == model.id
    assert model_updated.num != model.num
    assert model_updated.num == 5


def test_database_delete(conn):
    model = ModelUpdateDelete()
    insert = ModelUpdateDelete.t.insert().values(model.dict())
    c = ModelUpdateDelete.c

    delete = ModelUpdateDelete.t.delete().where(c.id == model.id)
    query = ModelUpdateDelete.t.select()

    conn.execute(insert)
    results = conn.execute(query)
    data_before_delete = results.fetchall()
    conn.execute(delete)
    results = conn.execute(query)
    data_after_delete = results.fetchall()

    assert len(data_before_delete) == 1
    assert len(data_after_delete) == 0


class LeftTable(Base):
    a: str = "a"


class RightTable(Base):
    b: str = "b"
    left_id: UUID4 = Field(
        sa_args=[sa.ForeignKey(LeftTable.c.id)]  # type: ignore[attr-defined]
    )


def test_database_join_with_pydantic_model(conn):

    left = LeftTable()
    insert_left = LeftTable.t.insert().values(left.dict())
    conn.execute(insert_left)

    right = RightTable(left_id=left.id)
    insert_right = RightTable.t.insert().values(right.dict())
    conn.execute(insert_right)

    join = RightTable.t.join(
        LeftTable.t, LeftTable.c.id == RightTable.c.left_id
    )
    query = sa.select([*RightTable.c, *LeftTable.c]).select_from(join)

    result_join = conn.execute(query)
    data = result_join.fetchall()

    assert data[0][0] == right.id
    assert data[0][2] == left.id


def test_database_join_with_table(conn):

    left = LeftTable()
    insert_left = LeftTable.t.insert().values(left.dict())
    conn.execute(insert_left)

    right = RightTable(left_id=left.id)
    insert_right = RightTable.t.insert().values(right.dict())
    conn.execute(insert_right)

    join = RightTable.t.join(
        LeftTable.t, LeftTable.c.id == RightTable.c.left_id
    )
    query = sa.select([*RightTable.c, *LeftTable.c]).select_from(join)

    result_join = conn.execute(query)
    data = result_join.fetchall()

    assert data[0][0] == right.id
    assert data[0][2] == left.id


def test_inheritance_validatable():
    class A(Validatable, metadata=MetaData()):
        id: int = Field(default_factory=lambda: randint(0, 10))

    class B(A):
        pass

    assert B.t is None


def test_inheritance_basemodel():
    class A(BaseTable, metadata=MetaData()):
        id: int = Field(default_factory=lambda: randint(0, 10))

    class B(A):
        pass

    assert B.t is None



import re
from importlib.machinery import SourceFileLoader
from pathlib import Path

from setuptools import setup


def normalize(name):
    return re.sub(r"[-_.]+", "-", name).lower()


THIS_DIR = Path(__file__).resolve().parent
long_description = (THIS_DIR / "README.md").read_text()
version = SourceFileLoader("version", "validatable/version.py").load_module()

setup(
    name="validatable",
    version=version.VERSION,
    description="Validatable provides a single class definition for data"
    " validation and SQL table representation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Daniel França",
    author_email="dcruzf.py@gmail.com",
    url="https://github.com/dcruzf/validatable",
    packages=["validatable"],
    license="MIT",
    python_requires=">=3.6.1",
    install_requires=["pydantic>=1.8", "sqlalchemy>=1.3", "sqlalchemy2-stubs"],
    extras_require={"email": ["email-validator>=1.0.3"]},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Operating System :: POSIX :: Linux",
        "Environment :: Console",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Internet",
    ],
)



import datetime as dt
import enum
import ipaddress
from collections import deque
from decimal import Decimal
from functools import partial
from pathlib import Path
from typing import Callable
from uuid import UUID
from weakref import WeakKeyDictionary, WeakSet

from pydantic import UUID1, UUID3, UUID4, UUID5, parse_raw_as
from pydantic.fields import ModelField
from pydantic.networks import (
    AnyHttpUrl,
    AnyUrl,
    EmailStr,
    HttpUrl,
    IPvAnyAddress,
    IPvAnyInterface,
    IPvAnyNetwork,
    NameEmail,
    pretty_email_regex,
)
from pydantic.types import (
    ConstrainedBytes,
    ConstrainedDecimal,
    ConstrainedFloat,
    ConstrainedInt,
    ConstrainedNumberMeta,
    ConstrainedStr,
    Json,
    JsonMeta,
    JsonWrapper,
    NegativeFloat,
    NegativeInt,
    NonNegativeFloat,
    NonNegativeInt,
    NonPositiveFloat,
    NonPositiveInt,
    PositiveFloat,
    PositiveInt,
    StrictBool,
    StrictFloat,
    StrictInt,
)
from sqlalchemy import (
    BigInteger,
    Boolean,
    Date,
    DateTime,
    Enum,
    Float,
    Integer,
    Interval,
    LargeBinary,
    Numeric,
    SmallInteger,
    String,
    Time,
)

from .generic_types import GUID, AutoJson, AutoString, SLBigInteger
from .typing import get_type, typing_meta


class Dispatch:
    def __init__(self, base: Callable):
        self._base = base
        self._funcs: WeakKeyDictionary = WeakKeyDictionary()

    def dispatcher(self, func: Callable):
        self._dispatcher = func

    def register(self, *args):
        self._types = WeakSet(args)
        return self.wrapper

    def wrapper(self, func):
        for t in self._types:
            self._funcs[t] = func
        self._types.clear()

    def __call__(self, *args, **kwargs):

        return self._dispatcher(*args, dispatch=self, **kwargs)


@Dispatch
def get_sql_type(m: ModelField, *args, **kwargs):
    raise TypeError(
        "cannot infer sqlalchemy " "type for {}".format(m.outer_type_)
    )


@get_sql_type.dispatcher
def _(m: ModelField, *args, dispatch: Dispatch = None, **kwargs):

    if m.outer_type_ in dispatch._funcs:
        func = dispatch._funcs.get(m.outer_type_)
        return func(m, *args, **kwargs)

    meta_func = dispatch._funcs.get(type(m.outer_type_))

    if meta_func:
        func = meta_func(m, dispatch=dispatch)
        if func:
            return func(m, *args, **kwargs)

    return dispatch._base(m)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ METACLASS


@get_sql_type.register(  # type: ignore[no-redef]
    ConstrainedNumberMeta,
    JsonMeta,
    type,
)
def _(m: ModelField, *args, dispatch: Dispatch = None, **kwargs):
    type_ = m.outer_type_
    if len(type_.__mro__) > 1 and type_.__mro__[1] in dispatch._funcs:
        return dispatch._funcs[type_.__mro__[1]]


@get_sql_type.register(enum.EnumMeta)  # type: ignore[no-redef]
def _(m: ModelField, *args, dispatch: Dispatch = None, **kwargs):
    return dispatch._funcs[enum.Enum]


@get_sql_type.register(*typing_meta)  # type: ignore[no-redef]
def _(m: ModelField, *args, dispatch: Dispatch = None, **kwargs):
    type_ = get_type(m.outer_type_)
    if type_ in dispatch._funcs:
        return dispatch._funcs[type_]
    return None


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ TYPES


@get_sql_type.register(str)  # type: ignore[no-redef]
def _(m: ModelField, *args, **kwargs):
    return AutoString


@get_sql_type.register(ConstrainedStr)  # type: ignore[no-redef]
def _(m: ModelField, *args, **kwargs):
    type_ = m.outer_type_
    length = type_.max_length or type_.curtail_length
    return String(length)


@get_sql_type.register(EmailStr)  # type: ignore[no-redef]
def _(m: ModelField, *args, **kwargs):
    return String(320)


@get_sql_type.register(NameEmail)  # type: ignore[no-redef]
def _(m: ModelField, *args, **kwargs):
    return AutoString(
        deserializer=lambda x: NameEmail(
            *pretty_email_regex.fullmatch(x).groups()
        )
    )


@get_sql_type.register(Path)  # type: ignore[no-redef]
def _(m: ModelField, *args, **kwargs):
    return AutoString(deserializer=Path)


@get_sql_type.register(  # type: ignore[no-redef]
    UUID, UUID1, UUID3, UUID4, UUID5
)
def _(m: ModelField, *args, **kwargs):
    return GUID


@get_sql_type.register(dt.datetime)  # type: ignore[no-redef]
def _(m: ModelField, *args, **kwargs):
    return DateTime


@get_sql_type.register(dt.date)  # type: ignore[no-redef]
def _(m: ModelField, *args, **kwargs):
    return Date


@get_sql_type.register(dt.time)  # type: ignore[no-redef]
def _(m: ModelField, *args, **kwargs):
    return Time


@get_sql_type.register(dt.timedelta)  # type: ignore[no-redef]
def _(m: ModelField, *args, **kwargs):
    return Interval


@get_sql_type.register(  # type: ignore[no-redef]
    IPvAnyAddress, ipaddress.IPv4Address, ipaddress.IPv6Address
)
def _(m: ModelField, *args, **kwargs):
    # https://datatracker.ietf.org/doc/html/rfc1924
    # IPv6 addresses, being 128 bits long, need 32 characters to write in
    # the general case, if standard hex representation, is used, plus more
    # for any punctuation inserted (typically about another 7 characters,
    # or 39 characters total).
    return AutoString(length=39, deserializer=ipaddress.ip_address)


@get_sql_type.register(  # type: ignore[no-redef]
    IPvAnyNetwork, ipaddress.IPv4Network, ipaddress.IPv6Network
)
def _(m: ModelField, *args, **kwargs):
    return AutoString(length=43, deserializer=ipaddress.ip_network)


@get_sql_type.register(  # type: ignore[no-redef]
    IPvAnyInterface, ipaddress.IPv4Interface, ipaddress.IPv6Interface
)
def _(m: ModelField, *args, **kwargs):
    return AutoString(length=43, deserializer=ipaddress.ip_interface)


@get_sql_type.register(HttpUrl)  # type: ignore[no-redef]
def _(m: ModelField, *args, **kwargs):
    return AutoString(length=HttpUrl.max_length)


@get_sql_type.register(AnyUrl, AnyHttpUrl)  # type: ignore[no-redef]
def _(m: ModelField, *args, **kwargs):

    return AutoString(length=AnyUrl.max_length)


@get_sql_type.register(  # type: ignore[no-redef]
    int,
    PositiveInt,
    NegativeInt,
    NonPositiveInt,
    NonNegativeInt,
    StrictInt,
)
def _(m: ModelField, *args, **kwargs):
    # https://docs.sqlalchemy.org/en/14/dialects/sqlite.html
    return SLBigInteger


@get_sql_type.register(ConstrainedInt)  # type: ignore[no-redef]
def _(m: ModelField, *args, **kwargs):
    type_ = m.outer_type_
    min_value = type_.gt or type_.ge
    max_value = type_.lt or type_.le
    if not (min_value and max_value):
        return BigInteger

    if min_value >= -32768 and max_value <= 32767:
        return SmallInteger

    if min_value >= -2147483648 and max_value <= 2147483647:
        return Integer

    return BigInteger


@get_sql_type.register(  # type: ignore[no-redef]
    float,
    ConstrainedFloat,
    PositiveFloat,
    NegativeFloat,
    NonPositiveFloat,
    NonNegativeFloat,
    StrictFloat,
)
def _(m: ModelField, *args, **kwargs):

    return Float


@get_sql_type.register(Decimal)  # type: ignore[no-redef]
def _(m: ModelField, *args, **kwargs):

    return Numeric


@get_sql_type.register(ConstrainedDecimal)  # type: ignore[no-redef]
def _(m: ModelField, *args, **kwargs):
    type_ = m.outer_type_
    return Numeric(precision=type_.max_digits, scale=type_.decimal_places)


@get_sql_type.register(bytes)  # type: ignore[no-redef]
def _(m: ModelField, *args, **kwargs):
    return LargeBinary


@get_sql_type.register(ConstrainedBytes)  # type: ignore[no-redef]
def _(m: ModelField, *args, **kwargs):
    type_ = m.outer_type_
    if type_.max_length:
        return LargeBinary(type_.max_length)
    return LargeBinary


@get_sql_type.register(enum.Enum)  # type: ignore[no-redef]
def _(m: ModelField, *args, **kwargs):
    type_ = m.outer_type_

    return Enum(type_)


@get_sql_type.register(JsonWrapper, Json)  # type: ignore[no-redef]
def _(m: ModelField, *args, **kwargs):
    return AutoJson


@get_sql_type.register(bool, StrictBool)  # type: ignore[no-redef]
def _(m: ModelField, *args, **kwargs):
    return Boolean


@get_sql_type.register(  # type: ignore[no-redef]
    list,
    set,
    tuple,
    deque,
    # Tuple,  # List, Set,
)
def _(m: ModelField, *args, **kwargs):
    loads = partial(parse_raw_as, m.outer_type_)
    return AutoJson(deserializer=loads, python_type=m.outer_type_)



import json
import uuid
from functools import partial
from typing import Any

import sqlalchemy as sa
from pydantic.json import pydantic_encoder


class SLBigInteger(sa.types.TypeDecorator):

    cache_ok = True
    impl = sa.BigInteger

    def load_dialect_impl(self, dialect):
        if dialect.name == "sqlite":
            return dialect.type_descriptor(sa.dialects.sqlite.INTEGER)
        else:
            return dialect.type_descriptor(sa.BigInteger)


class GUID(sa.types.TypeDecorator):
    """Platform-independent GUID type.
    Uses PostgreSQL's UUID type, otherwise uses
    BINARY(16), to store UUID.
    """

    cache_ok = True
    impl = sa.types.BINARY

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return dialect.type_descriptor(
                sa.dialects.postgresql.UUID(as_uuid=True)
            )
        else:
            return dialect.type_descriptor(sa.types.BINARY(16))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        if dialect.name == "postgresql":
            return value
        else:
            return value.bytes

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == "postgresql":
            return value
        else:
            return uuid.UUID(bytes=value)

    @property
    def python_type(self):
        return uuid.UUID


class AutoString(sa.types.TypeDecorator):
    """Platform-independent String type.
    Uses str(value) to bind parameter type.
    """

    cache_ok = True
    impl = sa.types.String
    length = 512
    serializer = str

    @staticmethod
    def deserializer(v):
        return v

    def __init__(
        self,
        length=None,
        collation=None,
        convert_unicode=False,
        unicode_error=None,
        _warn_on_bytestring=False,
        _expect_unicode=False,
        serializer=None,
        deserializer=None,
    ):
        self.serializer = serializer or self.serializer
        self.deserializer = deserializer or self.deserializer

        super().__init__(
            length=length,
            collation=collation,
            convert_unicode=convert_unicode,
            unicode_error=unicode_error,
            _warn_on_bytestring=_warn_on_bytestring,
            _expect_unicode=_expect_unicode,
        )

    def load_dialect_impl(self, dialect):
        if dialect.name in ("mysql", "mariadb"):
            return dialect.type_descriptor(sa.types.String(self.length))
        return dialect.type_descriptor(sa.types.String)

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        if isinstance(value, str):
            return value
        else:
            return self.serializer(value)

    def process_result_value(self, value, dialect):
        return self.deserializer(value)

    @property
    def python_type(self):
        return str


dumps = partial(json.dumps, default=pydantic_encoder)


class AutoJson(sa.types.TypeDecorator):
    """Json type with serialization"""

    cache_ok = True
    impl = sa.types.JSON

    def __init__(
        self,
        serializer=dumps,
        deserializer=lambda x: x,
        python_type=Any,
        none_as_null=False,
    ):
        self.serializer = serializer
        self.deserializer = deserializer
        self._python_type = python_type

        super().__init__(none_as_null=none_as_null)

    def load_dialect_impl(self, dialect):
        return dialect.type_descriptor(self.impl)

    def process_bind_param(self, value, dialect):
        return self.serializer(value)

    def process_result_value(self, value, dialect):
        return self.deserializer(value)

    @property
    def python_type(self):
        return self._python_type



import datetime as dt
import enum
import ipaddress
import numbers
import pathlib
import uuid
from decimal import Decimal
from ipaddress import (
    IPv4Address,
    IPv4Interface,
    IPv4Network,
    IPv6Address,
    IPv6Interface,
    IPv6Network,
)

import pytest
import sqlalchemy as sa
from pydantic import (
    UUID1,
    UUID3,
    UUID4,
    UUID5,
    AnyHttpUrl,
    AnyUrl,
    BaseModel,
    EmailStr,
    Field,
    HttpUrl,
    IPvAnyAddress,
    IPvAnyInterface,
    IPvAnyNetwork,
    NameEmail,
    StrictBool,
    conbytes,
    condecimal,
    confloat,
    conint,
    constr,
    stricturl,
)

from validatable.generic_types import GUID, AutoString, SLBigInteger
from validatable.inference import get_column


class CaseEnum(enum.Enum):
    a: int = 1
    b: int = 2
    c: str = "c"


MAX_LENGTH: int = 10


class ModelCase(BaseModel):

    id: int = Field(sa_primary_key=True)
    python_uuid: uuid.UUID
    uuid1: UUID1
    uuid3: UUID3
    uuid4: UUID4
    uuid5: UUID5
    python_int: int
    con_int: conint(strict=True)  # type: ignore[valid-type]
    python_float: float
    con_float: confloat(strict=True)  # type: ignore[valid-type]
    python_decimal: Decimal
    con_decimal: condecimal(max_digits=10)  # type: ignore[valid-type]
    python_str: str
    python_str_max: str = Field(max_length=MAX_LENGTH)
    con_str: constr(max_length=MAX_LENGTH)  # type: ignore[valid-type]
    con_str_curtain: constr(  # type: ignore[valid-type]
        curtail_length=MAX_LENGTH
    )
    con_str_curtain_max: constr(  # type: ignore[valid-type]
        curtail_length=MAX_LENGTH, max_length=2 * MAX_LENGTH
    )
    email_str: EmailStr
    name_email: NameEmail
    python_bytes: bytes
    python_bytes_max: bytes = Field(max_length=MAX_LENGTH)
    con_bytes: conbytes(max_length=MAX_LENGTH)  # type: ignore[valid-type]
    con_bytes_no_max: conbytes()  # type: ignore[valid-type]
    python_path: pathlib.Path
    ipv4: IPv4Address
    ipv4i: IPv4Interface
    ipv4n: IPv4Network
    ipv6: IPv6Address
    ipv6i: IPv6Interface
    ipv6n: IPv6Network
    ipvany: IPvAnyAddress
    ipvanyi: IPvAnyInterface
    ipvanyn: IPvAnyNetwork
    http_url: HttpUrl
    any_http_url: AnyHttpUrl
    any_url: AnyUrl
    strict_url: stricturl()  # type: ignore[valid-type]
    dt_datetime: dt.datetime
    dt_date: dt.date
    dt_time: dt.time
    dt_timedelta: dt.timedelta
    enum_field: CaseEnum
    name: int = Field(sa_column=sa.Column("not_valid", sa.Integer))

    sa_type: Decimal = Field(sa_type=sa.DECIMAL(precision=10))
    sa_column: str = Field(
        sa_column=sa.Column(sa.String(255), nullable=False), max_length=255
    )
    pbool: bool
    strict_bool: StrictBool
    bigint: conint()  # type: ignore[valid-type]
    bigint_after: conint(  # type: ignore[valid-type]
        gt=-2147483647, lt=2147483648
    )
    sqlint: conint(gt=-2147483648, lt=2147483647)  # type: ignore[valid-type]
    sqlsmallint: conint(gt=-32768, lt=32767)  # type: ignore[valid-type]


def test_column_declaration_with_invalid_name():

    with pytest.raises(ValueError):
        field = ModelCase.__fields__.get("name")
        get_column(field)


def test_get_column_python_int():
    """
    WHEN called with int
    THEN the SqlAlchemy type is BigInteger
    """
    field = ModelCase.__fields__.get("python_int")
    col = get_column(field)
    assert col.type.__class__ == SLBigInteger


def test_get_column_con_int():
    """
    WHEN called with ConstrainedIntValue
    THEN the SqlAlchemy type is BigInteger
    """
    con_int = ModelCase.__fields__.get("con_int")
    col = get_column(con_int)
    assert col.type.__class__ == sa.BigInteger


def test_get_column_python_float():
    """
    WHEN called with float
    THEN the SqlAlchemy type is Float
    """
    python_float = ModelCase.__fields__.get("python_float")
    col = get_column(python_float)
    assert col.type.__class__ == sa.Float


def test_get_column_con_float():
    """
    WHEN called with ConstrainedFloatValue
    THEN the SqlAlchemy type is Float
    """
    con_float = ModelCase.__fields__.get("con_float")
    col = get_column(con_float)
    assert col.type.__class__ == sa.Float


def test_get_column_python_decimal():
    """
    WHEN called with Decimal
    THEN the SqlAlchemy type is Decimal
    """
    python_decimal = ModelCase.__fields__.get("python_decimal")
    col = get_column(python_decimal)
    assert col.type.__class__ == sa.Numeric


def test_get_column_con_decimal():
    """
    WHEN called with ConstrainedDecimalValue
    THEN the SqlAlchemy type is Decimal
    """
    con_decimal = ModelCase.__fields__.get("con_decimal")
    col = get_column(con_decimal)
    assert col.type.__class__ == sa.Numeric


def test_get_column_python_str():
    """
    WHEN called with str
    THEN the SqlAlchemy type is Text
    """
    python_str = ModelCase.__fields__.get("python_str")
    col = get_column(python_str)
    assert col.type.__class__ == AutoString


def test_get_column_python_str_field_max_length():
    """
    WHEN called with str
    THEN the SqlAlchemy type is Text
    """
    python_str = ModelCase.__fields__.get("python_str_max")
    col = get_column(python_str)
    assert col.type.__class__ == sa.String
    assert col.type.length == MAX_LENGTH


def test_get_column_con_str():
    """
    WHEN called with ConstrainedStringValue with max_length=MAX_LENGTH
    THEN the SqlAlchemy type is String(MAX_LENGTH)
    """
    con_str = ModelCase.__fields__.get("con_str")
    col = get_column(con_str)
    assert col.type.__class__ == sa.String
    assert col.type.length == MAX_LENGTH


def test_get_column_con_str_curtain():
    """
    WHEN called with ConstrainedStringValue with max_length=MAX_LENGTH
    THEN the SqlAlchemy type is String(MAX_LENGTH)
    """
    con_str = ModelCase.__fields__.get("con_str_curtain")
    col = get_column(con_str)
    assert col.type.__class__ == sa.String
    assert col.type.length == MAX_LENGTH


def test_get_column_con_str_curtain_and_max_length():
    """
    WHEN called with ConstrainedStringValue with max_length=2*MAX_LENGTH
    THEN the SqlAlchemy type is String(2*MAX_LENGTH)
    """
    con_str = ModelCase.__fields__.get("con_str_curtain_max")
    col = get_column(con_str)
    assert col.type.__class__ == sa.String
    assert col.type.length == 2 * MAX_LENGTH


def test_get_column_email_str():
    """
    WHEN called with ConstrainedStringValue with max_length=MAX_LENGTH
    THEN the SqlAlchemy type is String(MAX_LENGTH)
    """
    email_str = ModelCase.__fields__.get("email_str")
    col = get_column(email_str)
    assert col.type.__class__ == sa.String
    assert col.type.length == 320


def test_get_column_name_email():
    """
    WHEN called with ConstrainedStringValue with max_length=MAX_LENGTH
    THEN the SqlAlchemy type is String(MAX_LENGTH)
    """
    name_email = ModelCase.__fields__.get("name_email")
    col = get_column(name_email)
    assert col.type.__class__ == AutoString


@pytest.mark.parametrize("field", ["con_bytes_no_max", "python_bytes"])
def test_get_column_python_bytes(field):
    """
    WHEN called with bytes
    THEN the SqlAlchemy type is LargeBinary
    """
    python_bytes = ModelCase.__fields__.get(field)
    col = get_column(python_bytes)
    assert col.type.__class__ == sa.LargeBinary


def test_get_column_python_bytes_field_max_length():
    """
    WHEN called with ConstrainedBytesValue
    THEN the SqlAlchemy type is LargeBinary
    """
    con_bytes = ModelCase.__fields__.get("python_bytes_max")
    col = get_column(con_bytes)
    assert col.type.__class__ == sa.LargeBinary
    assert col.type.length == MAX_LENGTH


def test_get_column_con_bytes():
    """
    WHEN called with ConstrainedBytesValue
    THEN the SqlAlchemy type is LargeBinary
    """
    con_bytes = ModelCase.__fields__.get("con_bytes")
    col = get_column(con_bytes)
    assert col.type.__class__ == sa.LargeBinary
    assert col.type.length == MAX_LENGTH


def test_get_column_python_path():
    """
    WHEN called with Path
    THEN the SqlAlchemy type is Text
    """
    python_path = ModelCase.__fields__.get("python_path")
    col = get_column(python_path)
    assert col.type.__class__ == AutoString


def test_get_column_sa_col():
    """
    WHEN called with sa_col
    THEN the SqlAlchemy Column is returned
    """
    sa_column = ModelCase.__fields__.get("sa_column")
    col = get_column(sa_column)
    assert col.type.__class__ == sa.String
    assert col.type.length == 255
    assert not col.nullable


def test_get_column_sa_type():
    """
    WHEN called with sa_type
    THEN the SqlAlchemy Column with type is returned
    """
    sa_type = ModelCase.__fields__.get("sa_type")
    col = get_column(sa_type)
    assert col.type.__class__ == sa.DECIMAL
    assert col.type.precision == 10


@pytest.mark.parametrize(
    ("field", "length"),
    (
        (
            ("ipv4", 31),
            ("ipv4i", 31),
            ("ipv4n", 31),
            ("ipv6", 43),
            ("ipv6i", 43),
            ("ipv6n", 43),
            ("ipvany", 43),
            ("ipvanyi", 43),
            ("ipvanyn", 43),
        )
    ),
)
def test_get_column_network_types(field: str, length: int):
    """
    WHEN called with networktype
    THEN the SqlAlchemy type is String(length)
    """
    network_type = ModelCase.__fields__.get(field)
    if not network_type:
        raise KeyError("invalid field name")
    col = get_column(network_type)
    assert col.type.__class__ == AutoString


@pytest.mark.parametrize(
    ("field", "sa_type"),
    (
        ("dt_datetime", sa.DateTime),
        ("dt_date", sa.Date),
        ("dt_time", sa.Time),
        ("dt_timedelta", sa.Interval),
    ),
)
def test_get_column_datetimes(field: str, sa_type):
    """
    WHEN called with datetime types
    THEN get correct SqlAlchemy type
    """
    dt_field = ModelCase.__fields__.get(field)
    col = get_column(dt_field)  # type: ignore
    assert col.type.__class__ == sa_type


def test_get_column_enum():
    """
    WHEN called with enum.Enum
    THEN returns correct SqlAlchemy type
    """
    field = ModelCase.__fields__.get("enum_field")
    col = get_column(field)
    assert col.type.__class__ == sa.Enum
    assert col.type.python_type == CaseEnum


@pytest.mark.parametrize(
    "field", ("python_uuid", "uuid1", "uuid3", "uuid4", "uuid5")
)
def test_get_column_uuid(field: str):
    """
    WHEN called with UUID
    THEN returns GUID
    """
    m = ModelCase.__fields__.get(field)
    col = get_column(m)  # type: ignore
    assert col.type.__class__ == GUID
    assert col.type.python_type == uuid.UUID


@pytest.mark.parametrize(
    "field", ("http_url", "any_http_url", "strict_url", "any_url")
)
def test_get_column_urls(field: str):
    m = ModelCase.__fields__.get(field)
    col = get_column(m)  # type: ignore
    assert col.type.__class__ == AutoString


@pytest.mark.parametrize("field", ("pbool", "strict_bool"))
def test_get_column_bool(field: str):
    m = ModelCase.__fields__.get(field)
    col = get_column(m)  # type: ignore
    assert col.type.__class__ == sa.Boolean


@pytest.mark.parametrize(
    ["field", "sql_type"],
    (
        ("bigint", sa.BigInteger),
        ("bigint_after", sa.BigInteger),
        ("sqlint", sa.Integer),
        ("sqlsmallint", sa.SmallInteger),
    ),
)
def test_get_column_sql_ints(field: str, sql_type):
    m = ModelCase.__fields__.get(field)
    col = get_column(m)  # type: ignore
    assert col.type.__class__ == sql_type


@pytest.mark.parametrize("T", (numbers.Real, ipaddress._IPAddressBase, object))
def test_invalid_type(T):
    class ModelCase(BaseModel):
        field: T

        class Config:
            arbitrary_types_allowed = True

    field = ModelCase.__fields__.get("field")
    with pytest.raises(TypeError):
        get_column(field)



from pydantic.class_validators import root_validator, validator
from pydantic.networks import (
    AnyHttpUrl,
    AnyUrl,
    EmailStr,
    HttpUrl,
    IPvAnyAddress,
    IPvAnyInterface,
    IPvAnyNetwork,
    NameEmail,
    PostgresDsn,
    RedisDsn,
    stricturl,
    validate_email,
)
from pydantic.types import (
    UUID1,
    UUID3,
    UUID4,
    UUID5,
    ByteSize,
    ConstrainedBytes,
    ConstrainedDecimal,
    ConstrainedFloat,
    ConstrainedInt,
    ConstrainedList,
    ConstrainedSet,
    ConstrainedStr,
    DirectoryPath,
    FilePath,
    Json,
    JsonWrapper,
    NegativeFloat,
    NegativeInt,
    NoneBytes,
    NoneStr,
    NoneStrBytes,
    NonNegativeFloat,
    NonNegativeInt,
    NonPositiveFloat,
    NonPositiveInt,
    PaymentCardNumber,
    PositiveFloat,
    PositiveInt,
    PyObject,
    SecretBytes,
    SecretStr,
    StrBytes,
    StrictBool,
    StrictBytes,
    StrictFloat,
    StrictInt,
    StrictStr,
    conbytes,
    condecimal,
    confloat,
    conint,
    conlist,
    conset,
    constr,
)
from sqlalchemy import ForeignKey, MetaData

from .engine import create_engine
from .fields import Field
from .main import BaseTable, Validatable
from .version import VERSION

__all__ = [
    "Validatable",
    "BaseTable",
    "Field",
    "__version__",
    # sqlalchemy
    "create_engine",
    "ForeignKey",
    "MetaData",
    # pydantic.class_validators
    "root_validator",
    "validator",
    # pydantic.networks
    "AnyUrl",
    "AnyHttpUrl",
    "HttpUrl",
    "stricturl",
    "EmailStr",
    "NameEmail",
    "IPvAnyAddress",
    "IPvAnyInterface",
    "IPvAnyNetwork",
    "PostgresDsn",
    "RedisDsn",
    "validate_email",
    # pydantic.types
    "NoneStr",
    "NoneBytes",
    "StrBytes",
    "NoneStrBytes",
    "StrictStr",
    "ConstrainedBytes",
    "conbytes",
    "ConstrainedList",
    "conlist",
    "ConstrainedSet",
    "conset",
    "ConstrainedStr",
    "constr",
    "PyObject",
    "ConstrainedInt",
    "conint",
    "PositiveInt",
    "NegativeInt",
    "NonNegativeInt",
    "NonPositiveInt",
    "ConstrainedFloat",
    "confloat",
    "PositiveFloat",
    "NegativeFloat",
    "NonNegativeFloat",
    "NonPositiveFloat",
    "ConstrainedDecimal",
    "condecimal",
    "UUID1",
    "UUID3",
    "UUID4",
    "UUID5",
    "FilePath",
    "DirectoryPath",
    "Json",
    "JsonWrapper",
    "SecretStr",
    "SecretBytes",
    "StrictBool",
    "StrictBytes",
    "StrictInt",
    "StrictFloat",
    "PaymentCardNumber",
    "ByteSize",
]
__version__ = VERSION



from datetime import datetime
from uuid import uuid4

import pytest
import sqlalchemy as sa
from pydantic import BaseModel, Json

from validatable import UUID4, BaseTable, Field, MetaData, create_engine


class JsonCase(BaseModel):
    id: int = 1
    name: str = "json case"
    created_ts: datetime = datetime.now()


metadata = MetaData()


def test_json_type():
    class TableJson(BaseTable, metadata=metadata):
        id: UUID4 = Field(default_factory=uuid4, sa_primary_key=True)
        json_field: Json

    assert isinstance(TableJson.__sa_table__, sa.Table)


@pytest.mark.parametrize(
    "valid_json",
    [
        b'{"name":"John", "age":30, "car":null}',
        b'{"employees":["John", "Anna", "Peter"]}',
        b"[1, 2, 3]",
        JsonCase().json(),
    ],
    ids=[0, 1, 2, 3],
)
def test_json_type_db(valid_json):
    metadata = MetaData()

    class TableJson(BaseTable, metadata=metadata):
        id: UUID4 = Field(default_factory=uuid4, sa_primary_key=True)
        data: Json

    model = TableJson(data=valid_json)

    engine = create_engine("sqlite:///:memory:")

    metadata.create_all(engine)
    with engine.connect() as conn:
        insert = TableJson.t.insert().values(model.dict())
        conn.execute(insert)
        result = conn.execute(TableJson.t.select())
        data = result.fetchone()
        m = TableJson.parse_obj(data)

    assert model == m


def test_json_type_parameterised():
    class TableJsonWrapper(BaseTable, metadata=metadata):
        id: UUID4 = Field(default_factory=uuid4, sa_primary_key=True)
        json_field: Json[JsonCase]

    assert isinstance(TableJsonWrapper.t, sa.Table)


def test_json_type_parameterised_db():
    metadata = MetaData()

    class TableJsonWrapper(BaseTable, metadata=metadata):
        id: UUID4 = Field(default_factory=uuid4, sa_primary_key=True)
        json_field: Json[JsonCase]

    model = TableJsonWrapper(json_field=JsonCase().json())

    engine = create_engine("sqlite:///:memory:")

    metadata.create_all(engine)
    with engine.connect() as conn:
        insert = TableJsonWrapper.t.insert().values(model.dict())
        conn.execute(insert)
        result = conn.execute(TableJsonWrapper.t.select())
        data = result.fetchone()
        m = TableJsonWrapper.parse_obj(data)

    assert model == m
